#include "mainwin.h"
#include <time.h>

#pragma comment(lib, "Crypt32.lib")

HCRYPTPROV hCryptProv = 0;
HCRYPTKEY hKey = 0;
HCRYPTHASH hHash = 0;
char buf[32] = "01STARTIN";
char letter[64];

extern Settings ProgramSettings;

void CryptCleanUp(void)
{
	if(hHash) CryptDestroyHash(hHash);

    //   .
    if(hKey) CryptDestroyKey(hKey);

    //   .
    if(hCryptProv) CryptReleaseContext(hCryptProv, 0);
}

BOOL CryptInit()
{

    //--------------------------------------------------------------------
    //     .

    if(!CryptAcquireContext(&hCryptProv, NULL, MS_ENHANCED_PROV, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT)) 
    {
        PrintLog(L"Error during CryptAcquireContext! %u", GetLastError());
	    return FALSE;
    }
    //--------------------------------------------------------------------
    //    .

    if(!CryptCreateHash(hCryptProv, CALG_MD5, 0, 0, &hHash)) 
    {
        PrintLog(L"Error during CryptCreateHash! %u", GetLastError());
	    return FALSE;
    }
    //--------------------------------------------------------------------
    //   .

	if(!CryptHashData(hHash, (BYTE *)ProgramSettings.CMDPass, (DWORD)wcslen(ProgramSettings.CMDPass), 0)) 
    {
        PrintLog(L"Error during CryptHashData! %u", GetLastError());
	    return FALSE;
    }
    //--------------------------------------------------------------------
    //   ,   ,   .

    if(!CryptDeriveKey(hCryptProv, CALG_3DES, hHash, NULL, &hKey)) 
    {
        PrintLog(L"Error during CryptDeriveKey! %u", GetLastError());
	    return FALSE;
    }

	__int64 id = time(0) - (GetCurrentThreadId() << 16) + GetCurrentProcessId();

	if(!CryptGenRandom(hCryptProv, 8, (BYTE *)&id)) 
    {
        PrintLog(L"Error during CryptGenRandom! %u", GetLastError());
	    return FALSE;
    }
	return TRUE;
}


// buf - data; dwSz - in size buf, out size letter ; letter - encrypted message; 

char *crypt_3des(BYTE *buf, DWORD dwSz)
{
	PBYTE CryptData = NULL;
	CryptData = (BYTE *)malloc(dwSz + 16);
    if(!CryptData)
    { 
        PrintLog(L"Out of memory for CryptData."); 
	    return FALSE;
    }

	if(!CryptGenRandom(hCryptProv, 8, CryptData))
    {
        PrintLog(L"Error during CryptGenRandom! %u", GetLastError());
	    return FALSE;
    }

	if(!CryptSetKeyParam(hKey, KP_IV, CryptData, 0))
    {
        PrintLog(L"Error during CryptSetKeyParam!");
	    return FALSE;
    }

	memcpy_s(CryptData + 8, dwSz + 8, buf, dwSz);

	if(!CryptEncrypt(hKey, NULL, TRUE, 0, CryptData + 8, &dwSz, dwSz + 8)) //
	{
		PrintLog(L"Error during CryptEncrypt! %u", GetLastError());
	}

	DWORD len;
	dwSz = dwSz + 8;
	if(0 == CryptBinaryToString((BYTE *)CryptData, dwSz, CRYPT_STRING_BASE64 | CRYPT_STRING_NOCR, NULL, &len))
	{
		PrintLog(L"Error during CryptBinaryToString! %u", GetLastError());
	}
	char *letter = (char *)malloc(len);
	if(!letter)
    { 
        PrintLog(L"Out of memory for letter."); 
	    return FALSE;
    }

	if(0 == CryptBinaryToStringA((BYTE *)CryptData, dwSz, CRYPT_STRING_BASE64 | CRYPT_STRING_NOCR, letter, &len))
	{
		PrintLog(L"Error during CryptBinaryToString! %u", GetLastError());
	}

	if(letter[len-1] == 10)
	{
		letter[len-1] = 0;
		dwSz = len-1;
	}

    // Free memory. 
    if(CryptData) 
    {
        free(CryptData); 
    }
	return letter;
} 

BYTE *decrypt_3des(char *letter, DWORD *dwSz)
{
	if(!CryptStringToBinaryA(letter, NULL, CRYPT_STRING_BASE64, NULL, dwSz, NULL, NULL)) 
	{
		PrintLog(L"Error during CryptStringToBinary! %u", GetLastError());
	}

	BYTE *buf = (BYTE *)malloc(*dwSz);
	if(!buf)
    { 
        PrintLog(L"Out of memory for buf."); 
	    return FALSE;
    }

	if(!CryptStringToBinaryA(letter, NULL, CRYPT_STRING_BASE64, buf, dwSz, NULL, NULL)) 
	{
		PrintLog(L"Error during CryptStringToBinary! %u", GetLastError());
	}

	if(!CryptSetKeyParam(hKey, KP_IV, buf, 0))
    {
        PrintLog(L"Error during CryptSetKeyParam!");
		if(buf) free(buf);
	    return FALSE;
    }
	*dwSz = *dwSz-8;
	BYTE *decrypted = (BYTE *)malloc(*dwSz);
	if(!decrypted)
    { 
        PrintLog(L"Out of memory for decrypted."); 
		if(buf) free(buf);
	    return FALSE;
    }

	memcpy_s(decrypted, *dwSz, buf + 8, *dwSz);

	if(buf) free(buf);

	if(!CryptDecrypt(hKey, NULL, TRUE, 0, decrypted, dwSz)) //
	{
		PrintLog(L"Error during CryptDecrypt! %u", GetLastError());
		if(decrypted) free(decrypted);
	    return FALSE;
	}

//	buf[dwSz] = 0;
    return decrypted;
} 
